package com.gic.udesk.service.outer.impl;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.gic.api.base.commons.ServiceResponse;
import com.gic.udesk.constant.UdeskSharedConstant;
import com.gic.udesk.dto.*;
import com.gic.udesk.entity.TabUdeskMessage;
import com.gic.udesk.entity.TabUdeskTapdRel;
import com.gic.udesk.entity.TabUdeskTicket;
import com.gic.udesk.service.*;
import com.gic.udesk.utils.UdeskConstant;
import com.gic.udesk.utils.UdeskHttpClientUtil;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import javax.annotation.Resource;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

@Service("tapdMessageApiService")
public class TapdMessageApiServiceImpl implements TapdMessageApiService {
    private static Logger logger = LogManager.getLogger(TapdMessageApiServiceImpl.class);

    @Resource(name = "udeskTapdRelMap")
    private Map<String, String> udeskTapdRelMap;
    @Autowired
    private UdeskTicketService udeskTicketService;
    @Autowired
    private UdeskMessageService udeskMessageService;
    @Autowired
    private UdeskBaseApiService udeskBaseApiService;
    @Autowired
    private UdeskTapdRelService udeskTapdRelService;

    /**
     * tapd缺陷接口
     */
    private static final String BUGS = "/bugs";
    /**
     * udesk工单接口
     */
    private static final String UDESK_UPDATE_TICKET = "/tickets/";
    /**
     * 客服回复
     */
    private static final String AGENT_REPLY = "/tickets/agent_reply";
    private static final String NOT_HANDLE = "NotHandle";
    private static final String TAPD_FIELD_NAME_PRIORITY = "priority";
    private static final String TAPD_FIELD_NAME_SEVERITY = "severity";

    @Override
    public ServiceResponse<String> handleTapdMessage(String tapdMessageStr) {
        //接收tapd消息数据
        TapdMessageDTO tapdMessageDTO = JSON.parseObject(tapdMessageStr, TapdMessageDTO.class);
        String event = tapdMessageDTO.getEvent();
        String handleResult = null;
        if (UdeskSharedConstant.TAPD_BUG_UPDATE_EVENT.equals(event)) {
            //tapd更新消息
            logger.info("TAPD发送了缺陷更新事件,缺陷id:{},变更字段:{}", tapdMessageDTO.getId(), tapdMessageDTO.getChange_fields());
            handleResult = handleTapdBugUpdate(tapdMessageDTO);
        } else if (UdeskSharedConstant.TAPD_BUG_DELETE_EVENT.equals(event)) {
            logger.info("TAPD发送了缺陷删除事件,缺陷id:{} ", tapdMessageDTO.getId());
            handleResult = handleTapdBugDelete(tapdMessageDTO);
        }
        return ServiceResponse.success(handleResult);
    }

    private String handleTapdBugUpdate(TapdMessageDTO tapdMessage) {
        // 查询TapdBug在中间件中记录
        List<TabUdeskTicket> dataBaseTicketList = this.getDataBaseTicket(tapdMessage);
        if (CollectionUtils.isEmpty(dataBaseTicketList)) {
            //没有建立关联 结束
            return NOT_HANDLE;
        }
        TabUdeskTicket dataBaseTicket = dataBaseTicketList.get(0);
        // 根据变更字段，来检查是否需要更新
        boolean needUpdate = this.checkChangeFields(tapdMessage.getChange_fields());
        if (!needUpdate) {
            return NOT_HANDLE;
        }
        //工单ID
        tapdMessage.setUdeskId(dataBaseTicket.getId().toString());
        TabUdeskMessage tabTapdMessage = this.getTapdMessageTemplate(tapdMessage,
                UdeskSharedConstant.TAPD_BUG_UPDATE_EVENT);
        Integer msgId = udeskMessageService.save(tabTapdMessage);
        logger.info("消息{}保存成功", tabTapdMessage);
        // 获取缺陷最新的内容
        TapdBugDTO latestBugFromTAPD = this.getLatestBugFromTAPD(tapdMessage);
        if (latestBugFromTAPD == null) {
            return msgId.toString();
        }
        // 数据转换一下. priorityEN,取值: "",high-高 ..
        String priorityEN = this.ensureFieldValueIsEn(latestBugFromTAPD.getPriority(), TAPD_FIELD_NAME_PRIORITY,
                tapdMessage);
        if (priorityEN == null) {
            return msgId.toString();
        }
        String priorityCN = udeskTapdRelMap.get(priorityEN + UdeskConstant.MAP_UDESK_NAME_SUFFIX);
        // severityEN,取值: "",advice-建议
        String severityEN = this.ensureFieldValueIsEn(latestBugFromTAPD.getSeverity(), TAPD_FIELD_NAME_SEVERITY,
                tapdMessage);
        if (severityEN == null) {
            return msgId.toString();
        }
        String severityCN = udeskTapdRelMap.get(severityEN + UdeskConstant.MAP_UDESK_NAME_SUFFIX);
        // 是否要更新中间件的数据库. 根据数据库表设计,数据库中ticket的status不会为null,priority、severity可能为null.
        if ((dataBaseTicket.getStatusTapd().equals(latestBugFromTAPD.getStatus()))
                && ((dataBaseTicket.getPriority() == null ? "" : dataBaseTicket.getPriority())
                        .equals(priorityCN == null ? "" : priorityCN))
                && ((dataBaseTicket.getSeverity() == null ? "" : dataBaseTicket.getSeverity())
                        .equals((severityCN == null ? "" : severityCN)))) {
            // 如果tapd最新的数据和数据库中的数据相同就不操作了,再完成消息处理
            tabTapdMessage.setStatus(UdeskConstant.DELETED);
            udeskMessageService.update(tabTapdMessage);
            return NOT_HANDLE;
        }
        // 同步Udesk的工单 1.拼装数据
        JSONObject updateTicketRequestBody = new JSONObject();
        JSONObject udeskTicketUpdate = new JSONObject();
        JSONObject customFields = new JSONObject();
        udeskTicketUpdate.put("status_id",
                udeskTapdRelMap.get(latestBugFromTAPD.getStatus() + UdeskConstant.MAP_UDESK_ID_VALUE_SUFFIX));
        udeskTicketUpdate.put("priority_id", udeskTapdRelMap.get(priorityEN + UdeskConstant.MAP_UDESK_ID_VALUE_SUFFIX));
        // key:value,value为null的话,不会被序列化.   为了防止误操作,如果TAPD的缺陷的严重程度为空,并不会去更新udesk中的严重程度.
        customFields.put(UdeskConstant.UDESK_FIELD_SEVERITY,
                udeskTapdRelMap.get(severityEN + UdeskConstant.MAP_UDESK_ID_VALUE_SUFFIX));
        udeskTicketUpdate.put("custom_fields", customFields);
        updateTicketRequestBody.put("ticket", udeskTicketUpdate);
        // 更新udesk的工单
        boolean updateUdeskTicketSuccess = this.updateUdeskTicket(tapdMessage, updateTicketRequestBody);
        if (!updateUdeskTicketSuccess) {
            return msgId.toString();
        }
        // 更新数据库中的数据  注: 放在udesk操作完成之后
        dataBaseTicket
                .setStatusEn(udeskTapdRelMap.get(latestBugFromTAPD.getStatus() + UdeskConstant.MAP_UDESK_NAME_SUFFIX));
        dataBaseTicket.setPriority(priorityCN);
        dataBaseTicket.setSeverity(severityCN);
        dataBaseTicket.setStatusTapd(latestBugFromTAPD.getStatus());
        udeskTicketService.update(dataBaseTicket);
        //消息处理完成
        tabTapdMessage.setStatus(UdeskConstant.DELETED);
        udeskMessageService.update(tabTapdMessage);
        return msgId.toString();
    }

    private String handleTapdBugDelete(TapdMessageDTO tapdMessage) {
        // 查询TapdBug在中间件中记录
        List<TabUdeskTicket> dataBaseTicketList = this.getDataBaseTicket(tapdMessage);
        if (CollectionUtils.isEmpty(dataBaseTicketList)) {
            //没有建立关联 结束
            return NOT_HANDLE;
        }
        TabUdeskTicket dataBaseTicket = dataBaseTicketList.get(0);
        // 保存消息
        tapdMessage.setUdeskId(dataBaseTicket.getId().toString());
        TabUdeskMessage tabTapdMessage = this.getTapdMessageTemplate(tapdMessage,
                UdeskSharedConstant.TAPD_BUG_DELETE_EVENT);
        Integer msgId = udeskMessageService.save(tabTapdMessage);
        logger.info("消息{}保存成功", tabTapdMessage);
        //在Udesk工单下添加缺陷删除消息
        UdeskAgentReplyDTO reply = new UdeskAgentReplyDTO();
        reply.setAgent_id(Integer.valueOf(udeskTapdRelMap.get(UdeskConstant.UDESK_DEFAULT_AGENT)));
        reply.setFunction_type(UdeskConstant.INTERNAL);
        reply.setId(Integer.valueOf(tapdMessage.getUdeskId()));
        reply.setNew_reply("缺陷Id:" + dataBaseTicket.getBugId() + " ,在TAPD已删除");
        //在udesk工单下面添加客服回复
        boolean addReplySuccess = this.addAgentInternalReplyInUdesk(tapdMessage, reply);
        if (!addReplySuccess) {
            return msgId.toString();
        }
        //更新数据库中ticket状态 注: 放在udesk操作完成之后
        dataBaseTicket.setStatus(UdeskConstant.DELETED);
        udeskTicketService.update(dataBaseTicket);
        //消息处理完成
        tabTapdMessage.setStatus(UdeskConstant.DELETED);
        udeskMessageService.update(tabTapdMessage);
        return msgId.toString();
    }

    /**
     * 根据tapdMessage在数据库中查找对应的工单
     *
     * @param tapdMessage tapd传来的message，当然是经过删减的
     * @return 查询结果
     */
    private List<TabUdeskTicket> getDataBaseTicket(TapdMessageDTO tapdMessage) {
        TabUdeskTicket params = new TabUdeskTicket();
        params.setBugId(tapdMessage.getId());
        List<TabUdeskTicket> list = udeskTicketService.list(params);
        return list;
    }

    /**
     * 检查变更字段,判断是否需要同步.
     * 当TAPD的 status-状态,priority-优先级,severity-严重程度 发生改变时，需要更新
     * @param changeFields 变更了的字段  e.g:  status,severity
     * @return 是否需要同步
     */
    private boolean checkChangeFields(String changeFields) {
        if (StringUtils.isBlank(changeFields)) {
            return false;
        }
        String[] changeFieldsArray = changeFields.split(",");
        for (String fieldName : changeFieldsArray) {
            // 当TAPD的 status-状态,priority-优先级,severity-严重程度 发生改变时，需要更新
            if ("status".equals(fieldName) || TAPD_FIELD_NAME_PRIORITY.equals(fieldName)
                    || TAPD_FIELD_NAME_SEVERITY.equals(fieldName)) {
                return true;
            }
        }
        return false;
    }

    /**
     * 根据tapd传来的message来转换为保存在数据库中的message模板
     *
     * @param tapdMessage tapd传来的message，当然是经过删减的
     * @param bugEvent    缺陷事件,e.g: bug::update,bug::delete
     * @return 存储在数据库中的消息
     */
    private TabUdeskMessage getTapdMessageTemplate(TapdMessageDTO tapdMessage, String bugEvent) {
        TabUdeskMessage tabTapdMessage = new TabUdeskMessage();
        tabTapdMessage.setBugId(tapdMessage.getId());
        tabTapdMessage.setTicketId(tapdMessage.getUdeskId());
        //消息来源 udesk/tapd
        tabTapdMessage.setSource(UdeskSharedConstant.TAPD_SOURCE);
        //消息主题 bug
        tabTapdMessage.setTopic(UdeskSharedConstant.TAPD_MESSAGE_TOPIC);
        //事件类型 update / delete
        tabTapdMessage.setEvent(bugEvent);
        //消息数据
        tabTapdMessage.setMsgBody(JSON.toJSONString(tapdMessage));
        tabTapdMessage.setStatus(UdeskConstant.UNDELETED);
        return tabTapdMessage;
    }

    /**
     * 从TAPD中获取缺陷最新的详情
     *
     * @param tapdMessage tapd的消息
     * @return TapdBugDTO
     */
    private TapdBugDTO getLatestBugFromTAPD(TapdMessageDTO tapdMessage) {
        Map<String, Object> queryBugParamsMap = new HashMap<>(4);
        queryBugParamsMap.put("id", tapdMessage.getId());
        queryBugParamsMap.put("workspace_id", tapdMessage.getWorkspace_id());
        logger.info("获取缺陷详情,请求内容:{}", queryBugParamsMap);
        String tapdApiResponseStr = UdeskHttpClientUtil.getTapdJsonFromParam(BUGS, queryBugParamsMap);
        logger.info("获取缺陷详情,请求返回:{}", tapdApiResponseStr);
        TapdRespBugDTO tapdRespBug = JSON.parseObject(tapdApiResponseStr, TapdRespBugDTO.class);
        if (tapdRespBug == null || UdeskConstant.TAPD_RESP_SUCCESS_STATUS != tapdRespBug.getStatus()) {
            String exceptionRecord = "查询TAPD缺陷失败;workSpaceId:" + tapdMessage.getWorkspace_id() + ";bugId:"
                    + tapdMessage.getId() + ";ticketId:" + tapdMessage.getUdeskId() + ";请求返回:" + tapdApiResponseStr;
            logger.info(exceptionRecord);
            udeskBaseApiService.saveTapdException(exceptionRecord);
            return null;
        }
        return tapdRespBug.getData().getBug();
    }

    /**
     * 获取tapd缺陷详情时，优先级(priority)和严重程度(serverityt)的值有时是中文、有时是英文。如果是传来的是中文，便将中文值转换为对应的英文值。
     *
     * @param fieldValue 字段值
     * @param fieldName  字段名
     * @return 对应的英文值或者是""; null的话就是有问题了.
     */
    private String ensureFieldValueIsEn(String fieldValue, String fieldName, TapdMessageDTO tapdMessage) {
        // fieldValue是""是由正常流程;是null，就是异常了走下面的return null;
        if ("".equals(fieldValue)) {
            return fieldValue;
        }
        // 如果在map里找得到映射就说明它是英文的,不用再查数据库找转换为中文了。
        if (StringUtils.isNotBlank(udeskTapdRelMap.get(fieldValue + UdeskConstant.MAP_UDESK_NAME_SUFFIX))) {
            return fieldValue;
        }
        TabUdeskTapdRel udeskTapdRel = null;
        if (TAPD_FIELD_NAME_PRIORITY.equals(fieldName)) {
            // 7对应优先级
            TabUdeskTapdRel params = new TabUdeskTapdRel();
            params.setRelType(7);
            params.setTapdName(fieldValue);
            udeskTapdRel = udeskTapdRelService.list(params).get(0);
        } else if (TAPD_FIELD_NAME_SEVERITY.equals(fieldName)) {
            // 4 对应严重程度
            TabUdeskTapdRel params = new TabUdeskTapdRel();
            params.setRelType(4);
            params.setUdeskName(fieldValue);
            udeskTapdRel = udeskTapdRelService.list(params).get(0);
        }
        if (udeskTapdRel == null) {
            String exceptionRecord = "转化字段值时,在缓存和数据库中都找不到对应的转换值;" + fieldName + ":" + fieldValue + ";workSpaceId:"
                    + tapdMessage.getWorkspace_id() + ";bugId:" + tapdMessage.getId() + ";ticketId:"
                    + tapdMessage.getUdeskId();
            logger.info(exceptionRecord);
            udeskBaseApiService.saveTapdException(exceptionRecord);
            return null;
        }
        return udeskTapdRel.getTapdId();
    }

    /**
     * 更新udesk的工单
     *
     * @param tapdMessage             tapd消息,包含ticketId、workSpaceId、bugId
     * @param updateTicketRequestBody udesk ticket的更新内容
     * @return 是否成功
     */
    private boolean updateUdeskTicket(TapdMessageDTO tapdMessage, JSONObject updateTicketRequestBody) {
        String updateTicketStr = updateTicketRequestBody.toJSONString();
        logger.info("更新udesk工单,工单Id:{},更新内容为:{}", tapdMessage.getUdeskId(), updateTicketStr);
        String updateTicketRespStr = UdeskHttpClientUtil
                .putUdeskJsonFromJson(UDESK_UPDATE_TICKET + tapdMessage.getUdeskId(), updateTicketStr);
        logger.info("更新udesk工单,工单返回为:{}", updateTicketRespStr);
        UdeskTicketDetailDTO updateDetail = JSON.parseObject(updateTicketRespStr, UdeskTicketDetailDTO.class);
        if (updateDetail == null || updateDetail.getCode() != UdeskConstant.UDESK_SUCCESS_CODE) {
            String exceptionRecord = "更新工单失败;workSpaceId:" + tapdMessage.getWorkspace_id() + ";bugId:"
                    + tapdMessage.getId() + ";ticketId:" + tapdMessage.getUdeskId() + ";请求返回:" + updateTicketRespStr;
            logger.info(exceptionRecord);
            udeskBaseApiService.saveUdeskException(exceptionRecord);
            return false;
        }
        return true;
    }

    /**
     * 在Udesk工单下添加客服回复
     *
     * @param tapdMessage 来自tapd的消息
     * @param reply       添加回复的内容
     * @return 是否成功
     */
    private boolean addAgentInternalReplyInUdesk(TapdMessageDTO tapdMessage, UdeskAgentReplyDTO reply) {
        String replyStr = JSON.toJSONString(reply);
        logger.info("回复udesk工单bug链接入参为{}", replyStr);
        String replyRespStr = UdeskHttpClientUtil.postUdeskJsonFromJson(AGENT_REPLY, replyStr);
        logger.info("回复udesk工单bug链接返回为{}", replyRespStr);
        UdeskRespReplyDTO replyDetail = JSON.parseObject(replyRespStr, UdeskRespReplyDTO.class);
        if (replyDetail == null || UdeskConstant.UDESK_SUCCESS_CODE != replyDetail.getCode()) {
            String exceptionRecord = "在udesk工单下添加缺陷删除消息失败;workSpaceId:" + tapdMessage.getWorkspace_id() + ";bugId:"
                    + tapdMessage.getId() + ";ticketId:" + tapdMessage.getUdeskId() + ";请求返回:" + replyRespStr;
            logger.info(exceptionRecord);
            udeskBaseApiService.saveUdeskException(exceptionRecord);
            return false;
        }
        return true;
    }
}
